#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
NBE-E4210 Exercise 6 EEG visualization
Created on 2017-11-28
Last edited on 2019-11-27

@author: Marko Havu
"""

import numpy as np
import mne
import matplotlib.pyplot as plt

# EEG measurement, RT file (name, number of blocks, number of stimuli in block)
measurements = [('group1', ('rt20191125110638.dat', 2 * 15, 5)),
                ('group2', ('rt20191126171858.dat', 2 * 15, 6))]

# Delay of the audio subsystem in the stimulus PC (milliseconds) 
AUDIO_DELAY = 200
# Stimulus codes
STIM_CODES = {'visual': 1, 'aural': 2}
def stim_code_index(key):
    by_value = sorted([(v, k) for k, v in STIM_CODES.items()])
    return [i for i, (v, k) in enumerate(by_value) if k == key]

class ReactionTimes:
    """Reaction time measurement"""
    def __init__(self, rt, n_blocks, n_stimuli):
        """Create a ReactionTimes object from a list of reaction times (rt),
           grouping them into a given number (n_blocks) of blocks, and a given
           number of stimuli (n_stimuli) per block"""
        self.stimuli = rt[:, 0]
        rt = np.resize(rt, (n_blocks, n_stimuli, 2))
        rt = rt[rt.argsort(axis=0)[:,0,0], :, 1]  # unshuffle
        rt.resize(2, n_blocks // 2, n_stimuli)
        self.visual = rt[stim_code_index('visual')]
        self.aural = rt[stim_code_index('aural')]
        self.n_blocks = n_blocks
        self.n_stimuli = n_stimuli
    
    @staticmethod
    def stim_convert(stim):
        """Convert stimulus type in text file to a number"""
        return STIM_CODES[stim.decode('utf-8')]
    
    @staticmethod
    def rt_convert(rt):
        """Convert reaction time in text file to milliseconds"""
        return round(1000 * float(rt.strip()))
    
    @classmethod
    def load(cls, filename, n_blocks, n_stimuli):
        """Read reaction times from file"""
        rt_data = np.loadtxt(filename, delimiter=',', dtype='int',
                             converters={0: ReactionTimes.stim_convert,
                                         1: ReactionTimes.rt_convert})
        return cls(rt_data, n_blocks, n_stimuli)
    
    def droplist(self, stimulus, keep='shortest'):
        """Create a list of event indices to drop, keeping just the
           stimulus set specified by parameter 'keep' as follows:
               'shortest': set with the shortest mean reaction time
               'least_variance': set with the least variation in the RT"""
        rt = getattr(self, stimulus)
        if keep == 'shortest':
            pos = np.argmin(np.mean(rt, axis=2))
        elif keep == 'least_variance':
            pos = np.argmin(np.std(rt, axis=2))
        else:
            errmsg = 'Droplist keeping %s not implemented yet' % keep
            raise NotImplementedError(errmsg)
        indices = [i * self.n_stimuli + j
                   for i in range(self.n_blocks // 2) if i != pos
                   for j in range(self.n_stimuli)]
        return np.array(indices)

def read_eeg(filename):
    """Read EEG data and filter it"""
    raw = mne.io.read_raw_brainvision(filename, preload=True)
    montage = mne.channels.read_montage('standard_1020')
    raw.set_montage(montage)
    raw.set_eeg_reference('average', projection=True)
    # Band-pass filter from 1 Hz to 40 Hz
    raw.filter(1., 40., n_jobs=2, fir_design='firwin')
    return raw

def visualize(measurement, rt_data, plot_raw=False, plot_epochs=False):
    """Visualize a measurement"""
    # Add group and measurement to the plot titles
    v_title = measurement + '_visual'
    a_title = measurement + '_aural'
    
    # Read measurement data
    rt = ReactionTimes.load(*rt_data)
    eeg = read_eeg(measurement + '.vhdr')
    
    # Plot raw EEG data
    if plot_raw:
        eeg.plot(block=True)
    
    # Find marker events
    events = mne.find_events(eeg)
    assert len(events) == len(rt.stimuli)
    # Correct for audio delay by moving the event markers forward
    delay = round(AUDIO_DELAY / 1000 * eeg.info['sfreq'])
    events[np.where(events[:, 2] == STIM_CODES['aural']), 0] += delay
    
    # Get 1 s epochs around the markers
    v_epochs = mne.Epochs(eeg, events, event_id=STIM_CODES['visual'],
                          tmin=-0.2, tmax=0.8, proj=False)
    a_epochs = mne.Epochs(eeg, events, event_id=STIM_CODES['aural'],
                          tmin=-0.2, tmax=0.8, proj=False)
    # Manually drop epochs with eye blinks
    if measurement == 'group1':
        v_epochs.drop([0, 1, 2, 3, 5, 6, 7, 18, 29, 32, 33, 36, 46, 48, 51, 62,
                       72])
        a_epochs.drop([1, 5, 6, 10, 11, 15, 22, 28, 33, 35, 36, 39, 40, 49, 52,
                       56, 56, 59, 64, 67, 70, 71])
    elif measurement == 'group2':
        v_epochs.drop([3, 6, 9, 11, 46, 68, 83, 86])
        a_epochs.drop([7, 8, 18, 55, 71, 75, 89])
    else:
        pass
    # Alternatively, use only the selected series
    # (in this case, the one with least variance in reaction times)
    #v_epochs.drop(rt.droplist('visual', keep='least_variance'))
    #a_epochs.drop(rt.droplist('aural', keep='least_variance'))
    v_epochs.load_data()
    a_epochs.load_data()
    # Plot the 1 s epochs
    if plot_epochs:
        v_epochs.plot(title=v_title)
        a_epochs.plot(title=a_title)
    
    # Average evoked reponses
    v_epochs.average().plot()
    plt.savefig(v_title + '_butterfly.pdf', bbox_inches='tight')
    a_epochs.average().plot()
    plt.savefig(a_title + '_butterfly.pdf', bbox_inches='tight')
    t = np.arange(-0.025, 0.500, 0.025)
    v_epochs.average().plot_topomap(title=v_title, times=t[0:20])
    plt.savefig(v_title + '_topomap.pdf', bbox_inches='tight')
    a_epochs.average().plot_topomap(title=a_title, times=t[0:20])
    plt.savefig(a_title + '_topomap.pdf', bbox_inches='tight')


if __name__ == '__main__':
    for measurement in measurements:
        visualize(*measurement, plot_raw=False, plot_epochs=False)
